/**********************************************************************
 * File  : geometry.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 14/07/2007
 *********************************************************************/

#include "geometry.hpp"


bool ORect::IntersectsCircle(Vector2i center, float radius) const
{
	// early discard facile :
	// si le centre est en dehors de la boite qui englobe toute les rotations
	// et agrandie du rayon, alors le cercle est hors collisions.

	Rect bounding;
	bounding.w = (int)(2 * radius + 1.f) + rect.w + rect.h;  // +1.f pour eviter que le cast diminue le rayon reel.
	bounding.h = bounding.w;
	bounding.SetCenter(rect.GetCenter());

	if (!bounding.IsPointInside(center))  // pour avoir une chance d'etre en collision il faut passer ce test
		return false;
			
	// sinon le cercle est dehors, soit totalement, soit en partie en intersection avec une arete,
	// soit le centre est totalement compris dans le rectangle :
	
	UpdateEdgePt();

	float d;
	bool inside = true;
	Vector2f v;
	Vector2f pt[4];  // liste des projections
	Vector2f centerf = center;
	Vector2f orth;

	for (int edge = 0; edge < 4; ++edge)
	{
		orth = uedge[edge].GetOrtho();  // normale a l'arete
		v = point[edge] - centerf;
		d = v * orth;   // projection
		if (d < 0)
			inside = false;  // si toutes les projections sont positives, alors le centre est a l'interieur du rectangle
		pt[edge] = centerf + orth * d;
	}
	if (inside)
		return true;     // early accepted
			
	// ramenons nos "pt" dans les limites des aretes :
			
	for (int edge = 0; edge < 4; ++edge)
	{
		v = pt[edge] - point[edge];
		d = v * edges[edge];
		if (d > edges[edge].LenghtSq())  // on deborde
		{
			pt[edge] = point[(edge + 1) % 4];
		}
		else if (d < 0)
		{
			pt[edge] = point[edge];
		}
	}
			
	// testons si un des pt est dans le cercle, on accepte la collision :
			
	for (int edge = 0; edge < 4; ++edge)
	{
		if ((pt[edge] - centerf).LenghtSq() < radius * radius)
			return true;
	}
			
	return false;
}

bool ORect::IntersectsRay(float* t, const Vector2f& start, const Vector2f& direction) const
{
	UpdateEdgePt();
	
	Vector2f pt[4];
	float ts[4];
	bool ok[4];
	Vector2f orth;
	float d;
	float denom;
	for (int edge = 0; edge < 4; ++edge)
	{
		orth = uedge[edge].GetOrtho();
		d = (start - point[edge]) * orth;    // distance de "start" a l'arete
		denom = -direction * orth;
		if (Abs(denom) < 1e-4f) // rayon parallele a l'arete
			ok[edge] = false;
		else
		{
			ts[edge] = d / denom;
			if (ts[edge] < 0)   // collision derriere
				ok[edge] = false;
			else
			{
				ok[edge] = true;
				pt[edge] = start + direction * ts[edge];
			}
		}
	}
	// verifier si les points sont dans le champ des aretes
	Vector2f v;
	for (int edge = 0; edge < 4; ++edge)
	{
		if (ok[edge])
		{
			v = pt[edge] - point[edge];
			d = v * edges[edge];  // carre scalaire
			if (d > edges[edge].LenghtSq() || d < 0)  // deborde
			{
				ok[edge] = false; // cette arete ne collisionne pas
			}
		}
	}
	
	// trouver le t le plus petit
	float st = 1e37f;
	bool collide = false;
	for (int edge = 0; edge < 4; ++edge)
	{
		if (ok[edge])
		{
			if (ts[edge] < st)
			{
				st = ts[edge];
			}
			collide = true;
		}
	}
	if (collide)
		*t = st;  // renseigner le parametre pour l'user.
	
	return collide;
}

bool ORect::IntersectsORc(const ORect& r2) const
{
	Rect bound1, bound2;
	
	// premier discard : intersection des AABB
	
	bound1.w = rect.w + rect.h;
	bound1.h = bound1.w;
	bound1.SetCenter(rect.GetCenter());

	bound2.w = r2.rect.w + r2.rect.h;
	bound2.h = bound2.w;
	bound2.SetCenter(r2.rect.GetCenter());
	if (!bound1.Intersects(bound2))
	{
		return false;
	}
	
	UpdateEdgePt();
	r2.UpdateEdgePt();
	
	// second test : y'a t-il un point de l'une des boite au sein de l'autre ?
	for (int pt = 0; pt < 4; ++pt)
	{
		if (IsPtInside_pr(r2.point[pt]))  // de la boite 2 dans 1 ?
			return true;
	
		if (r2.IsPtInside_pr(point[pt]))  // de 1 dans 2 ?
			return true;
	}
	
	
	// sinon, peut etre les segments se croisent :
	for (int seg1 = 0; seg1 < 4; ++seg1)
	{
		for (int seg2 = 0; seg2 < 4; ++seg2)
		{
			if (SegmentsCollides(   point[seg1],    edges[seg1],
										r2.point[seg2], r2.edges[seg2]))
				return true;
		}
	}
	return false;  // si rien ne se croise ni aucun point ne s'entremele, les boites sont hors collisions
}

bool ORect::IsPtInside_pr(const Vector2f& a_point, bool noUpdate) const
{
	if (!noUpdate)
		UpdateEdgePt();
	
	Vector2f v;
	Vector2f orth;
	float d;

	for (int edge = 0; edge < 4; ++edge)
	{
		orth = uedge[edge].GetOrtho();  // n
		v = point[edge] - a_point;
		d = v * orth;
		if (d < 0)
			return false;
	}	
	return true;   // si toutes les projections sont positives
}

bool SegmentsCollides(const Vector2f& start1, const Vector2f& seg1, const Vector2f& start2, const Vector2f& seg2)
{
	Vector2f	dir1 = Normalize(seg1);
	Vector2f dir2 = Normalize(seg2);

	float d = (start2 - start1) * dir2.GetOrtho();  // distance entre point start1 et la droite 2
	float slope = dir1 * dir2.GetOrtho();  // ~0 si parallele. 1 ou -1 si orthogonaux

	if (Abs(slope) < 1e-5f)
		return false;

	float t = d / slope;
	if (t * t > seg1.LenghtSq() || t < 0)  // hors segment 1
		return false;

	Vector2f inter = start1 + dir1 * t;  // point d'intersection
	float proj = (inter - start2) * seg2;
	if (proj > seg2.LenghtSq() || proj < 0)  // hors segment 2
		return false;
	
	return true;  // si tout est passe jusque la, alors croisement.
}
