/**********************************************************************
 * File  : geometry.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 * Date  : 10/07/2007
 *********************************************************************/

#ifndef GEOM_NOVD_
#define GEOM_NOVD_ 1

#include <math.h>

template< typename T >
struct Vector2
{
		T x, y;
		
		Vector2(void) { }

		Vector2(T a_x, T a_y)
				: x(a_x),
				  y(a_y) { }
      
      // constructeur de casting :
      template< typename T2 >
      Vector2(const Vector2< T2 >& refT)
      {
         x = static_cast< T >(refT.x);
         y = static_cast< T >(refT.y);
      }
      
      // assignation cast :
      template< typename T2 >
      Vector2& operator = (const Vector2< T2 >& op2)
      {
         x = static_cast< T >(op2.x);
         y = static_cast< T >(op2.y);
         return *this;
      }
		
		Vector2 operator + (const Vector2& op2) const
		{
			return Vector2(x + op2.x, y + op2.y);
		}
			
		Vector2 operator - (const Vector2& op2) const
		{
			return Vector2(x - op2.x, y - op2.y);
		}
		
		Vector2 operator - (void) const  // - unaire
		{
			return Vector2(-x, -y);
		}

		T operator * (const Vector2& op2) const  // pdt scalaire
		{
			return x * op2.x + y * op2.y;
		}
		
		Vector2 operator * (T scal) const
		{
			return Vector2(x * scal, y * scal);
		}
		
		Vector2& operator += (const Vector2& op2)
		{
			x += op2.x;
			y += op2.y;
			return *this;
		}
		
		Vector2& operator -= (const Vector2& op2)
		{
			x -= op2.x;
			y -= op2.y;
			return *this;
		}

		Vector2& operator *= (T scal)
		{
			x *= scal;
			y *= scal;
			return *this;
		}
      
		bool operator == (const Vector2& op2)
		{
			return x == op2.x && y == op2.y;
		}

		bool operator != (const Vector2& op2)
		{
			return x != op2.x || y != op2.y;
		}
      
      T LenghtSq(void) const
      {
         return x * x + y * y;
      }
      
      Vector2 GetOrtho(void) const
      {
         return Vector2(-y, x);
      }
};

typedef Vector2< int > Vector2i;
typedef Vector2< float > Vector2f;

inline Vector2f Rotate(const Vector2f& vec, float angle)
{
	float fcos = cosf(angle);
	float fsin = sinf(angle);
	return Vector2f( vec.x * fcos + vec.y * fsin,
						 -vec.x * fsin + vec.y * fcos);
}

inline Vector2f Normalize(const Vector2f& in)
{
   const float rsq = 1.f / sqrtf(in.LenghtSq());  // reciprocal square root.
   return Vector2f(in.x * rsq, in.y * rsq);
}

inline float ToDegree(float angle)
{
	return angle * 57.295779524f;
}

inline float ToRadians(float angle)
{
	return angle * 0.017453293f;
}

template< typename T >
T Max(const T& v1, const T& v2)
{
	return (v1 > v2) ? v1 : v2;
}

template< typename T >
T Min(const T& v1, const T& v2)
{
	return (v1 < v2) ? v1 : v2;
}

template< typename T >
T Abs(const T& val)
{
	return Max(val, -val);
}

template< typename T >
T Sign(const T& val)
{
	return (val > 0) ? static_cast<T>(1) : static_cast<T>(-1);
}

inline float FltModulo(float val, const float modulo)
{
   if (val >= modulo || val < modulo)
   {
      int hm = (int)(val / modulo);  // combien de fois modulo dans val ?
      if (val < 0)
         hm -= 1;
      val -= hm * modulo;  // retour entre 0 et modulo.
   }
   return val;
}

struct Rect
{
		Vector2i pos;
		int w;
		int h;
		
		bool IsPointInside(const Vector2i& pt)
		{
			return pt.x >= pos.x && pt.x <= pos.x + w &&
				    pt.y >= pos.y && pt.y <= pos.y + h;
		}
		Vector2i GetCenter(void) const
		{
			return Vector2i(pos.x + (w >> 1), pos.y + (h >> 1));
		}
		void SetCenter(Vector2i center)
		{
			pos.x = center.x - w / 2;
			pos.y = center.y - h / 2;
		}
		
		bool Intersects(const Rect& r2) const
		{
			// vecteur reliant les centres :
			Vector2i axis = r2.GetCenter() - GetCenter();
			
			return Abs(axis.x) <= w / 2 + r2.w / 2 &&
				    Abs(axis.y) <= h / 2 + r2.h / 2;
		}
		
		bool ContainedIn(const Rect& r2)
		{
			return pos.x >= r2.pos.x && pos.x + w <= r2.pos.x + r2.w &&
				    pos.y >= r2.pos.y && pos.y + h <= r2.pos.y + r2.h;
		}
		
		void AddPoint(const Vector2i& pt)   // elargir la boite (si besoin) pour contenir ce point.
		{
			if (pos.x + w < pt.x)
				w = pt.x - pos.x;
			else if (pos.x > pt.x)
			{
				w += pos.x - pt.x;
				pos.x = pt.x;
			}
			if (pos.y + h < pt.y)
				h = pt.y - pos.y;
			else if (pos.y > pt.y)
			{
				h += pos.y - pt.y;
				pos.y = pt.y;
			}
		}
		
		Rect(void) { }
		
		Rect(int a_x, int a_y, int a_w, int a_h)
				: pos(a_x, a_y),
				  w(a_w),
				  h(a_h) { }
};

// oriented rectangle
struct ORect
{
      Rect rect;
      Vector2f dir;  // direction
		
				
		ORect(const Rect& cpy)
				: rect(cpy),
				  dir(1.f, 0.f) { }
		
		ORect(void) { }
		
		bool IntersectsRay(float* t, const Vector2f& start, const Vector2f& direction) const;
      
      bool IntersectsCircle(Vector2i center, float radius) const;
		
		bool IntersectsORc(const ORect& r2) const;
		
		bool IsPointInside(const Vector2f& a_point) const
		{
			return IsPtInside_pr(a_point, false);
		}
		
		Rect GetBoundingRect(void) const
		{
			Rect out;
			out.w = rect.w + rect.h;
			out.h = out.w;
			out.SetCenter(rect.GetCenter());
			return out;
		}
		
		ORect operator * (float factor) const
		{
			Vector2i oldCenter = rect.GetCenter();
			ORect ret;
			ret.rect.w = (int)(rect.w * factor);
			ret.rect.h = (int)(rect.h * factor);
			ret.rect.SetCenter(oldCenter);
			ret.dir = dir;
			return ret;
		}
		
		ORect& operator *= (float factor)
		{
			Vector2i oldCenter = rect.GetCenter();
			rect.w = (int)(rect.w * factor);
			rect.h = (int)(rect.h * factor);
			rect.SetCenter(oldCenter);
			return *this;
		}

		
	private:
		
		bool IsPtInside_pr(const Vector2f& point, bool noUpdate = true) const;

		inline void UpdateEdgePt(void) const;

		mutable Vector2f uedge[4];
		mutable Vector2f edges[4];
		mutable Vector2f point[4];
};

void ORect::UpdateEdgePt(void) const
{
	// construire les vecteurs aretes directrices de la boite :
         
	//          edge0              dir.GetOrtho()
	//       0 ------> 1         ^
	//       ^         |         |
	// edge3 |         | edge1   x-->  dir
	//       |         v
	//       3 <------ 2
	//          edge2
         
	// aretes unitaires :
	uedge[0] = dir;
	uedge[1] = -dir.GetOrtho();
	uedge[2] = -dir;
	uedge[3] = -uedge[1];
	// aretes :
	edges[0] = uedge[0] * rect.w;
	edges[1] = uedge[1] * rect.h;
	edges[2] = -edges[0];
	edges[3] = -edges[1];
	// on tourne le rectangle sur son centre pour le mettre en direction de "dir".
	// dir doit etre normalise !
	point[0] = rect.GetCenter();
	point[0] = point[0] + edges[2] * 0.5f + edges[3] * 0.5f;
	// les autres points :
	point[1] = point[0] + edges[0];
	point[2] = point[1] + edges[1];
	point[3] = point[2] + edges[2];
}


bool SegmentsCollides(const Vector2f& start1, const Vector2f& seg1, const Vector2f& start2, const Vector2f& seg2);

#endif
