/**********************************************************************
 * File  : smart_ptr.hpp
 * Author: Oddou Vivien - a.k.a. Lightness1024!
 *         mailto:vivien@isit.or.jp
 * Date  : 17/05/2007
 *********************************************************************/

/*
 *  -- WHAT IS A SMART POINTER ? --
 * 
 * this file declares and defines a class of smart pointer.
 * a smart pointer bechaves like a C-pointer. Except that when it goes out of scope,
 * or is destroyed, it triggers the destruction of the pointed object.
 * this smart pointer also has the convenient feature of reference counting.
 * if many smart pointers are copied, and therefore points to the same object,
 * then the destruction of one of them will not imply the destruction of the pointed object,
 * until there is only one pointer left.
 * A very good document for state-of-the-art programming of smart pointers can be found here:
 *    http://www.awprofessional.com/articles/article.asp?p=31529&seqNum=1&rl=1  
 */


#ifndef SMART_PTR_HPP_1024_
#define SMART_PTR_HPP_1024_ 1

#ifndef NULL
#define NULL 0
#endif

template< typename T >
class smart_ptr
{
protected:

	T* m_ptr;
	int* refcount;       // shared reference counter

	// note to implementor : beware not to call any Acquire-style function without having Released before.
	void Acquire(const smart_ptr& a_right)
	{
		m_ptr = a_right.m_ptr;
		refcount = a_right.refcount;
		++*refcount;
	}

	void Release(void)
	{
		if (m_ptr != NULL)
		{
			--*refcount;
			if (*refcount == 0)
			{
				delete m_ptr;
				delete refcount;
				m_ptr = NULL;
			}
		}
	}
	
public:

	// default construction
	smart_ptr(void) : m_ptr(NULL)
	{ /* ... */ }

	// creation from raw pointer :
	smart_ptr(T* a_object) : m_ptr(a_object),
	                         refcount(new int(1))
	{ /* ... */ }

	// creation from another smart_ptr :
	smart_ptr(const smart_ptr& a_2cpy)
	{
		Acquire(a_2cpy);
	}

	// deletion :
	~smart_ptr()
	{
		Release();
	}
	
	// copy from = operator
	const T& operator = (const smart_ptr& a_right)
	{
		Release();
		Acquire(a_right);
		return *m_ptr;
	}

	// access to data :
	T& operator * (void) const
	{
		return *m_ptr;
	}
   T* operator -> (void) const
   {
   	return m_ptr;
   }
   
   // raw access to pointer (warning, this is a leak)
   T* GetPtr(void) const
   {
   	return m_ptr;
   }
};

#endif
