/**********************************************************************
 * File  : spritemanager.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 09/07/2007
 *********************************************************************/

#include "spritemanager.hpp"

// definition des membres static

int SpriteManager::instances = 0;
SDL_Surface* SpriteManager::screen_surface = NULL;
int SpriteManager::resx;
int SpriteManager::resy;
int SpriteManager::scroll_x = 0;
int SpriteManager::scroll_y = 0;


// bliter un sprite
void SpriteManager::Draw(int id, int posX, int posY)
{
	Sprite& sprite = *(ap_sprites[id]);
	if (posX != -1)
	{
		sprite.posx = posX;
		sprite.posy = posY;
	}
	
	SDL_Rect rcSrc, rcDest;
	
	if (sprite.nbFrames > 0)  // sprite a animation
	{
		float tv = timer.Read() - lastTiming;  // interval de temps pour l'animation
		// pour animer, on avance dans les frames:
		sprite.animation += tv * sprite.speed;
		int frame = (int)sprite.animation;
		frame %= sprite.nbFrames;
		rcSrc.x = frame * sprite.surface->h;  // frame * largeur du carre = frame * hauteur de l'image
		rcSrc.y = 0;
		rcSrc.w = sprite.surface->h;          // w = hauteur (car le sprite est carre)
		rcSrc.h = sprite.surface->h;
	}
	else // sprite a rotation
	{
		// l'angle du sprite n'est jamais plus eloigne du vrai angle que de "delta" / 2.
		float angle = sprite.angle + sprite.angle_delta * 0.5f;
		int w = sprite.surface->w >> 2;  // divise par 4
		// determiner la case a lire dans le sprite selon l'angle
		if (angle < 180.f)  // premier cas, descendant.
		{
			if (angle < 90)
			{
				rcSrc.x = 0;
				rcSrc.y = (int)(angle / sprite.angle_delta);
			}
			else
			{
				rcSrc.y = (int)((angle - 90) / sprite.angle_delta);
				rcSrc.x = w;
			}
			rcSrc.y *= w;
		}
		else  // au dela de 180, c'est montant.
		{
			if (angle < 270)
			{
				rcSrc.x = w << 1;  // fois 2
				rcSrc.y = sprite.surface->h / w - 1 - (int)((angle - 180) / sprite.angle_delta);
			}
			else
			{
				rcSrc.y = sprite.surface->h / w - 1 - (int)((angle - 270) / sprite.angle_delta);
				if (rcSrc.y < 0)
					rcSrc.y = 0;   // possible car angle peut aller jusqu'a 360 + d/2
				rcSrc.x = w * 3;
			}
			rcSrc.y *= w;
		}
		rcSrc.w = w;
		rcSrc.h = w;
	}
	if (scroll_sensible)  // tenir compte du scrolling de l'ecran.
	{
		rcDest.x = sprite.posx - scroll_x;   // on pose le sprite a la position "coin en haut a gauche" du sprite.
		rcDest.y = sprite.posy - scroll_y;
	}
	else
	{
		rcDest.x = sprite.posx;
		rcDest.y = sprite.posy;
	}
	
	// w et h ne sont pas lus.

	SDL_BlitSurface(ap_sprites[id]->surface,
						 &rcSrc,
						 screen_surface,
						 &rcDest);
}

// charger un sprite
int SpriteManager::LoadSprite(const char* fname, SPRITE_PARAM_en parm)
{
	// ajouter un sprite:
	ap_sprites.Add(new Sprite);

	MapNameSurface::iterator it;
	it = surfaces_names.find(string(fname)); 
	if (it != surfaces_names.end())  // ce fichier a deja ete charge
	{
		ap_sprites.Back()->surface = it->second;  // on reference la surface existante
	}
	else
	{
		SDL_Surface* ptemp = SDL_LoadBMP(fname);  // sinon on la charge
		if (ptemp == NULL)
		{
			throw NoSuchFile_Or_LoadError();
		}
		else
		{
			ap_sprites.Back()->surface = SDL_DisplayFormat(ptemp);
			SDL_FreeSurface(ptemp);
			surfaces_names[fname] = ap_sprites.Back()->surface;
			int colorkey = SDL_MapRGB(ap_sprites.Back()->surface->format, 255, 0, 255);
			SDL_SetColorKey(ap_sprites.Back()->surface, SDL_SRCCOLORKEY | SDL_RLEACCEL, colorkey);
		}
	}
	ap_sprites.Back()->posx = 0;
	ap_sprites.Back()->posy = 0;
	ap_sprites.Back()->speed = 4.0f;

	if (parm == SP_ROTATED)
	{
		ap_sprites.Back()->nbFrames = 0;
		ap_sprites.Back()->angle_delta = 360.f / (16 * ap_sprites.Back()->surface->h / ap_sprites.Back()->surface->w);
		ap_sprites.Back()->angle = 0.f;
	}
	else
	{
		ap_sprites.Back()->animation = 0.f;
		ap_sprites.Back()->nbFrames = ap_sprites.Back()->surface->w / ap_sprites.Back()->surface->h;
	}

	return ap_sprites.GetUbound();
}

void SpriteManager::Update(UPDATE_PARAM_en opt)
{
	lastTiming = timer.Read();
	if (opt == UP_SCREEN_AND_TIME)
		SDL_Flip(screen_surface);
}

// initialiseur
void SpriteManager::Init(int resolutionX, int resolutionY, const char* title)
{
	if (instances == 0)
	{
		SDL_Init(SDL_INIT_VIDEO);
		SDL_WM_SetCaption(title, title);
		screen_surface = SDL_SetVideoMode(resolutionX, resolutionY, 0, SDL_HWSURFACE | SDL_DOUBLEBUF);
		if (screen_surface == NULL)
			throw CantCreateWindow();
		resx = resolutionX;
		resy = resolutionY;
	}
	
	scroll_sensible = true;
	timer.Start();
	++instances;
	ap_sprites.ReserveExtra(20);
}

SpriteManager::~SpriteManager()
{
	MapNameSurface::iterator it;
	for (it = surfaces_names.begin(); it != surfaces_names.end(); ++it)
	{
		SDL_FreeSurface(it->second);   // liberation des surfaces
	}
	if (--instances == 0)
	{
		SDL_Quit();
	}
}
