/**********************************************************************
 * File  : spritemanager.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 09/07/2007
 *********************************************************************/

#ifndef GAMECLASS_SOLARDIALES_HPP_
#define GAMECLASS_SOLARDIALES_HPP_ 1

#include <map>
#include <string>
#include <SDL/SDL.h>

#include "SecureArray.hpp"
#include "timer.hpp"
#include "smart_ptr.hpp"
#include "geometry.hpp"

using std::map;
using std::string;

struct Sprite
{
		int posx, posy;
		SDL_Surface* surface;
		int nbFrames;
		union
		{
				double animation;
				struct
				{
						float angle;
						float angle_delta;
				};
		};
		float speed;
};

typedef smart_ptr< Sprite > SpriteSmartPtr;

enum UPDATE_PARAM_en
{
	UP_SCREEN_AND_TIME,
	UP_TIME_ONLY
};

enum SPRITE_PARAM_en
{
	SP_ANIMATED,
	SP_ROTATED
};

class SpriteManager
{
	protected:

		struct NoSuchFile_Or_LoadError {};
		struct CantCreateWindow {};
		typedef map< string, SDL_Surface* > MapNameSurface;
		
		MapNameSurface surfaces_names;
		SecureArray< SpriteSmartPtr > ap_sprites;
		CTimer timer;
		float lastTiming;
		bool scroll_sensible;

		static SDL_Surface* screen_surface;
		static int instances;
		static int resx, resy;

		void Init(int resolutionX, int resolutionY, const char* title);

	public:
		
		static int scroll_x, scroll_y;  // shared
		
		int LoadSprite(const char* fname, SPRITE_PARAM_en parm = SP_ANIMATED);
		void Draw(int id, int posX = -1, int posY = -1);
		void DrawAll(unsigned id_from = 0)
		{
			for (int i = id_from; i < ap_sprites.GetElementsNumber(); ++i)
			{
				Draw(i);
			}
		}
		void Move(int id, int posX, int posY)
		{
			ap_sprites[id]->posx = posX;
			ap_sprites[id]->posy = posY;
		}
		void MoveRelative(int id, int displaceX, int displaceY)
		{
			ap_sprites[id]->posx += displaceX;
			ap_sprites[id]->posy += displaceY;
		}
		void SetSpeed(int id, float speed)
		{
			ap_sprites[id]->speed = speed;
		}
		void Rotate(int id, float angle)
		{
         angle = FltModulo(angle, 360.f);
			ap_sprites[id]->angle = angle;
		}
		int GetW(int id) const
		{
			if (ap_sprites[id]->nbFrames > 0)    // anime
				return ap_sprites[id]->surface->h;  // square
			else
				return ap_sprites[id]->surface->w >> 2;  // a rotation, 4 cadrans en colonne
		}
		int GetH(int id) const
		{
			return GetW(id);
		}
		static int GetResX(void)
		{
			return resx;
		}
		static int GetResY(void)
		{
			return resy;
		}
		void SetScrollSensible(bool ss)
		{
			scroll_sensible = ss;
		}

		void Update(UPDATE_PARAM_en opt = UP_SCREEN_AND_TIME);
		
		void ClearScreen(int color = 0)
		{
			SDL_FillRect(screen_surface, NULL, color);
		}

		// constructeurs
		SpriteManager(int resolutionX, int resolutionY, const char* title)
		{
			Init(resolutionX, resolutionY, title);
		}
		SpriteManager(void)
		{
			Init(640, 480, "noname");
		}

		~SpriteManager();
};


#endif
