/**********************************************************************
 * File  : terraingenerator.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 10/07/2007
 *********************************************************************/

#ifndef TER_GEN_NOVENDIALES_
#define TER_GEN_NOVENDIALES_ 1

#include "geometry.hpp"
#include "randclass.hpp"
#include "SecureArray.hpp"
#include "spritemanager.hpp"
#include <list>

using std::list;

typedef list< const char * > CharPtrList;

// enumerate all kind of object we can have in a map's setup :
// from bigger to smaller is better for procedural generation
enum KIND_OF_TERRAIN_OBJECTS_en
{
	KTO_WATER,
	KTO_ROCKS,
	KTO_TREE,
	KTO_BONUS,
	
	NB_KIND_OF_TERRAIN_OBJECTS
};

struct TerrainParameters
{
		int seed;
		
		// unit is / l'unite est :
		// nombre de repetitions du sprite de terrain
		// number of tiling repeats of background sprite
		int size_long;
		int size_height;
		
		// unit is :
		// 0 = no ; 1.0 = full coverage
		float density[NB_KIND_OF_TERRAIN_OBJECTS];
		
		// file name for ground tile :
		const char* terrain_sprite;
		
		// file name for external border wall sprite :
		const char* wall_sprite;

		// list of sprite's file name strings for each kind of terrain objects:
		CharPtrList sprites[NB_KIND_OF_TERRAIN_OBJECTS];
};

struct ORect;

struct TerrainObject
{
		Rect rect;
		int spriteId;
		bool highplane;
		
		TerrainObject(const Rect& a_rc, int spId, bool highZ)
				: rect(a_rc),
				  spriteId(spId),
				  highplane(highZ) { }

		virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction) = 0;
		virtual bool IntersectORect(const ORect& orc) = 0;
		virtual bool IntersectRect(const Rect& rc) = 0;

		virtual ~TerrainObject() { }
};

struct Tree : TerrainObject
{
		Tree(const Rect& a_rc, int spId)
				: TerrainObject(a_rc, spId, true) { }
      
      virtual bool IntersectRect(const Rect& rc)
      {
         return ORect(rc).IntersectsCircle(rect.GetCenter(), rect.w / 8);
      }
		virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction)
		{
			ORect orc(rect);
			orc *= 0.3f;
			return orc.IntersectsRay(t, start, direction);
		}
		virtual bool IntersectORect(const ORect& a_orc)
		{
			return a_orc.IntersectsCircle(rect.GetCenter(), rect.w / 8);
		}
};

struct Water : TerrainObject
{
		Water(const Rect& a_rc, int spId)
				: TerrainObject(a_rc, spId, false) { }
		
      virtual bool IntersectRect(const Rect& rc)
      {
         return ORect(rc).IntersectsCircle(rect.GetCenter(), rect.w / 2);
      }
		virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction)
		{
			ORect orc(rect);
			orc *= 0.8f;
			return orc.IntersectsRay(t, start, direction);
		}
		virtual bool IntersectORect(const ORect& a_orc)
		{
			return a_orc.IntersectsCircle(rect.GetCenter(), rect.w / 2);
		}
};

struct Rocks : TerrainObject
{
		Rocks(const Rect& a_rc, int spId)
				: TerrainObject(a_rc, spId, false) { }
		
      virtual bool IntersectRect(const Rect& rc)
      {
         return ORect(rect).IntersectsORc(ORect(rc));
      }
		virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction)
		{
			ORect orc(rect);
			return orc.IntersectsRay(t, start, direction);
		}
		virtual bool IntersectORect(const ORect& a_orc)
		{
			return ORect(rect).IntersectsORc(a_orc);
		}
};

struct Bonus : TerrainObject
{
		Bonus(const Rect& a_rc, int spId)
				: TerrainObject(a_rc, spId, false) { }
		
      virtual bool IntersectRect(const Rect& rc)
      {
         return ORect(rect).IntersectsORc(ORect(rc));
      }
		virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction)
		{
			ORect orc(rect);
			return orc.IntersectsRay(t, start, direction);
		}
		virtual bool IntersectORect(const ORect& a_orc)
		{
			return ORect(rect).IntersectsORc(a_orc);
		}
};

struct Weapon : Bonus
{
};

struct Repair : Bonus
{
};

struct ObjInstancier
{
		virtual TerrainObject* Create(const Rect&, int) = 0;
		virtual ~ObjInstancier() { }
};
		
template< class C >
struct Instancier : ObjInstancier
{
		C* Create(const Rect& rc, int id)
		{
			return new C(rc, id);
		}
};

template< typename Data_T >
struct BiNode
{
		BiNode* left;
		BiNode* right;

		Data_T data;
		
		bool IsLeaf(void) const
		{
			return left == NULL && right == NULL;
		}
		
		BiNode(void)
				: left(NULL),
				  right(NULL) { }
		
		BiNode(const Data_T& a_dt)
				: left(NULL),
				  right(NULL),
				  data(a_dt) { }

		~BiNode()
		{
			if (left)
				delete left;
			if (right)
				delete right;
		}
};

struct Sector
{
		struct FlagIndex
		{
				int index;           // reference to object stored at "index" in array "objects" of TM class.
				int flag;
				FlagIndex(void)
						: flag(-1) { }

				FlagIndex(int i, int f = -1)
						: index(i),
						  flag(f) { }
		};
		SecureArray< FlagIndex > contents;
		Rect rect;		
};

typedef BiNode< Sector > SectorNode;

class TerrainManager
{
		// space partitionning :
		SectorNode root;
		// scene container :
		SecureArray< TerrainObject* > objects;
		
		// water, terrain, rocks, bonus
		SpriteManager background;
		// les arbres doivent etre dessinnes au dessus des voitures:
		// trees must be drawn above cars:
		SpriteManager highground;
		
		bool loaded;
		Vector2i levelSize;
		int borderWidth;
		
		// tableau (persistant) pour passer a la fonction de detection des noeuds actives par un rectangle
		// pour le clipping, la detection des collisions, etc...
		SecureArray< SectorNode* > m_nodeList;
		int m_numNodes;
		int m_frame;
		
		// -- private methods --
		void BuildInDepth(SectorNode& node);

		CharPtrList::const_iterator GetRandomIt(const CharPtrList& a_list, int max, CRandom& rand) const;

		bool IsOverObjectFromNode(const Rect& rc, const SectorNode& node);
		bool IsRectIntersectingObjectOfNode(const Rect& rc, const SectorNode& node);
		void AddFromNode(int index, SectorNode& node);
		inline void InsertObject(TerrainObject* pobj);
		void BuildActivatedNodesList(const Rect& rc, SectorNode& node, SecureArray< SectorNode* >& nodeList, int& numNodes);
		
	public:
		
		// -- public interface --

		bool IsOverObject(const Rect& rc)
		{
			return IsOverObjectFromNode(rc, root);
		}
		
		void DrawBackGround(void);
		void DrawHighGround(void);

		void Update(UPDATE_PARAM_en p = UP_TIME_ONLY)
		{
			background.Update(p);
			highground.Update(p);
		}
		void Scroll(int x, int y)
		{
			SpriteManager::scroll_x = x;
			SpriteManager::scroll_y = y;
		}
		Vector2i GetSize(void)
		{
			return levelSize;
		}
		int GetBorderWidth(void)
		{
			return borderWidth;
		}
		
		void Generate(TerrainParameters& params);
		
		// collisions detection functions :
		bool WorldIntersectRect(const Rect& rc);
		bool WorldIntersectRay(float* t, const Vector2f& start, const Vector2f& direction, bool lowspan = true);
		bool WorldIntersectORect(const ORect& a_orc);
		
		
		TerrainManager(void)
				: loaded(false),
				  levelSize(0, 0),
				  borderWidth(0),
				  m_frame(0)
		{
			m_nodeList.Resize(32);
		}

		// destroy managed objects when time has come to die :
		~TerrainManager()
		{
			for (int i = 0; i < objects.GetElementsNumber(); ++i)
			{
				delete objects[i];
			}
		}
};

void TerrainManager::InsertObject(TerrainObject* pobj)
{
	// ajout a la collection :
	objects.Add(pobj);
	// puis ajout de l'indice dans les listes des feuilles qui intersectent :
	// add new index in leaves's lists that intersects :
	AddFromNode(objects.GetUbound(), root);
}

#endif
